from __future__ import annotations

import math

import pyarrow as pa
import pytest

from daft.datatype import DataType
from daft.series import Series
from tests.series import ARROW_FLOAT_TYPES, ARROW_INT_TYPES


@pytest.mark.parametrize("dtype", ARROW_FLOAT_TYPES)
def test_series_float_sorting(dtype) -> None:
    data = pa.array([5.0, 4.0, 1.0, None, 2.0, None, float("nan"), -float("nan"), float("inf"), -float("inf")])
    sorted_order = [-float("inf"), 1.0, 2.0, 4.0, 5.0, float("inf"), -float("nan"), float("nan"), None, None]
    s = Series.from_arrow(data.cast(dtype))
    s_sorted = s.sort()
    assert len(s_sorted) == len(s)
    assert s_sorted.datatype() == s.datatype()
    assert all(
        [
            (left == right) or (math.isnan(left) and math.isnan(right))
            for left, right in zip(s_sorted.to_pylist(), sorted_order)
        ]
    )

    s_argsorted = s.argsort()
    assert len(s_argsorted) == len(s)

    taken = s.take(s_argsorted)
    assert len(taken) == len(s)
    assert all(
        [
            (left == right) or (math.isnan(left) and math.isnan(right))
            for left, right in zip(taken.to_pylist(), sorted_order)
        ]
    )

    ## Descending
    s_sorted = s.sort(descending=True)
    assert len(s_sorted) == len(s)
    assert s_sorted.datatype() == s.datatype()
    assert all(
        [
            (left == right) or (math.isnan(left) and math.isnan(right))
            for left, right in zip(s_sorted.to_pylist(), sorted_order[::-1])
        ]
    )

    s_argsorted = s.argsort(descending=True)
    assert len(s_argsorted) == len(s)

    taken = s.take(s_argsorted)
    assert len(taken) == len(s)
    assert all(
        [
            (left == right) or (math.isnan(left) and math.isnan(right))
            for left, right in zip(taken.to_pylist(), sorted_order[::-1])
        ]
    )


@pytest.mark.parametrize("dtype", ARROW_INT_TYPES)
def test_series_int_sorting(dtype) -> None:
    data = pa.array([5, 4, 1, None, 2, None])
    sorted_order = [1, 2, 4, 5, None, None]
    s = Series.from_arrow(data.cast(dtype))
    s_sorted = s.sort()
    assert len(s_sorted) == len(s)
    assert s_sorted.datatype() == s.datatype()
    assert s_sorted.to_pylist() == sorted_order

    s_argsorted = s.argsort()
    assert len(s_argsorted) == len(s)

    taken = s.take(s_argsorted)
    assert len(taken) == len(s)
    assert taken.to_pylist() == sorted_order

    ## Descending
    s_sorted = s.sort(descending=True)
    assert len(s_sorted) == len(s)
    assert s_sorted.datatype() == s.datatype()
    assert s_sorted.to_pylist() == sorted_order[::-1]

    s_argsorted = s.argsort(descending=True)
    assert len(s_argsorted) == len(s)

    taken = s.take(s_argsorted)
    assert len(taken) == len(s)
    assert taken.to_pylist() == sorted_order[::-1]


def test_series_date_sorting() -> None:
    from datetime import date

    def date_maker(d):
        if d is None:
            return None
        return date(2023, 1, d)

    days = [5, 4, 1, None, 2, None]
    s = Series.from_pylist(list(map(date_maker, days)))
    sorted_order = list(map(date_maker, [1, 2, 4, 5, None, None]))
    s = s.cast(DataType.date())
    s_sorted = s.sort()
    assert len(s_sorted) == len(s)
    assert s_sorted.datatype() == s.datatype()
    assert s_sorted.to_pylist() == sorted_order

    s_argsorted = s.argsort()
    assert len(s_argsorted) == len(s)

    taken = s.take(s_argsorted)
    assert len(taken) == len(s)
    assert taken.to_pylist() == sorted_order

    ## Descending
    s_sorted = s.sort(descending=True)
    assert len(s_sorted) == len(s)
    assert s_sorted.datatype() == s.datatype()
    assert s_sorted.to_pylist() == sorted_order[::-1]

    s_argsorted = s.argsort(descending=True)
    assert len(s_argsorted) == len(s)

    taken = s.take(s_argsorted)
    assert len(taken) == len(s)
    assert taken.to_pylist() == sorted_order[::-1]


@pytest.mark.parametrize("timeunit", ["us", "ns"])
def test_series_time_sorting(timeunit) -> None:
    from datetime import time

    def time_maker(h, m, s, us):
        if us is None:
            return None
        return time(h, m, s, us)

    times = list(map(time_maker, [0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [5, 4, 1, None, 2, None]))
    s = Series.from_pylist(times)
    sorted_order = list(
        map(time_maker, [0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [1, 2, 4, 5, None, None])
    )
    s = s.cast(DataType.time(timeunit))
    s_sorted = s.sort()
    assert len(s_sorted) == len(s)
    assert s_sorted.datatype() == s.datatype()
    assert s_sorted.to_pylist() == sorted_order

    s_argsorted = s.argsort()
    assert len(s_argsorted) == len(s)

    taken = s.take(s_argsorted)
    assert len(taken) == len(s)
    assert taken.to_pylist() == sorted_order

    ## Descending
    s_sorted = s.sort(descending=True)
    assert len(s_sorted) == len(s)
    assert s_sorted.datatype() == s.datatype()
    assert s_sorted.to_pylist() == sorted_order[::-1]

    s_argsorted = s.argsort(descending=True)
    assert len(s_argsorted) == len(s)

    taken = s.take(s_argsorted)
    assert len(taken) == len(s)
    assert taken.to_pylist() == sorted_order[::-1]


def test_series_string_sorting() -> None:
    data = pa.array(["hi", "bye", "thai", None, "2", None, "h", "by"])
    sorted_order = ["2", "by", "bye", "h", "hi", "thai", None, None]
    s = Series.from_arrow(data)
    s_sorted = s.sort()
    assert len(s_sorted) == len(s)
    assert s_sorted.datatype() == s.datatype()
    assert s_sorted.to_pylist() == sorted_order

    s_argsorted = s.argsort()
    assert len(s_argsorted) == len(s)

    taken = s.take(s_argsorted)
    assert len(taken) == len(s)
    assert taken.to_pylist() == sorted_order

    ## Descending
    s_sorted = s.sort(descending=True)
    assert len(s_sorted) == len(s)
    assert s_sorted.datatype() == s.datatype()
    assert s_sorted.to_pylist() == sorted_order[::-1]

    s_argsorted = s.argsort(descending=True)
    assert len(s_argsorted) == len(s)

    taken = s.take(s_argsorted)
    assert len(taken) == len(s)
    assert taken.to_pylist() == sorted_order[::-1]


def test_series_boolean_sorting() -> None:
    data = pa.array([True, False, True, None, False])
    sorted_order = [False, False, True, True, None]
    s = Series.from_arrow(data)
    s_sorted = s.sort()
    assert len(s_sorted) == len(s)
    assert s_sorted.datatype() == s.datatype()
    assert s_sorted.to_pylist() == sorted_order

    s_argsorted = s.argsort()
    assert len(s_argsorted) == len(s)

    taken = s.take(s_argsorted)
    assert len(taken) == len(s)
    assert taken.to_pylist() == sorted_order

    ## Descending
    s_sorted = s.sort(descending=True)
    assert len(s_sorted) == len(s)
    assert s_sorted.datatype() == s.datatype()
    assert s_sorted.to_pylist() == sorted_order[::-1]

    s_argsorted = s.argsort(descending=True)
    assert len(s_argsorted) == len(s)

    taken = s.take(s_argsorted)
    assert len(taken) == len(s)
    assert taken.to_pylist() == sorted_order[::-1]



from typing import TYPE_CHECKING

from daft.lazy_import import LazyImport

if TYPE_CHECKING:
    import fsspec
    import numpy as np
    import pandas as pd
    import PIL.Image as pil_image
    import pyarrow as pa
    import pyarrow.csv as pacsv
    import pyarrow.dataset as pads
    import pyarrow.fs as pafs
    import pyarrow.json as pajson
    import pyarrow.parquet as pq
else:
    fsspec = LazyImport("fsspec")
    np = LazyImport("numpy")
    pd = LazyImport("pandas")
    pil_image = LazyImport("PIL.Image")
    pa = LazyImport("pyarrow")
    pacsv = LazyImport("pyarrow.csv")
    pads = LazyImport("pyarrow.dataset")
    pafs = LazyImport("pyarrow.fs")
    pajson = LazyImport("pyarrow.json")
    pq = LazyImport("pyarrow.parquet")

unity_catalog = LazyImport("daft.unity_catalog")



import pytest


@pytest.mark.parametrize("repartition_nparts", [1, 2, 5])
def test_pivot(make_df, repartition_nparts, with_morsel_size):
    daft_df = make_df(
        {
            "group": ["A", "A", "B", "B"],
            "pivot": [1, 2, 1, 2],
            "value": [10, 20, 30, 40],
        },
        repartition=repartition_nparts,
    )
    daft_df = daft_df.pivot(group_by="group", pivot_col="pivot", value_col="value", agg_fn="sum")

    expected = {
        "group": ["A", "B"],
        "1": [10, 30],
        "2": [20, 40],
    }

    assert daft_df.sort("group").to_pydict() == expected


@pytest.mark.parametrize("repartition_nparts", [1, 2, 5])
def test_pivot_with_col_names(make_df, repartition_nparts, with_morsel_size):
    daft_df = make_df(
        {
            "group": ["A", "A", "B", "B"],
            "pivot": [1, 2, 1, 2],
            "value": [10, 20, 30, 40],
        },
        repartition=repartition_nparts,
    )
    daft_df = daft_df.pivot(
        group_by="group",
        pivot_col="pivot",
        value_col="value",
        agg_fn="sum",
        names=["1", "2"],
    )

    expected = {
        "group": ["A", "B"],
        "1": [10, 30],
        "2": [20, 40],
    }

    assert daft_df.sort("group").to_pydict() == expected


@pytest.mark.parametrize("repartition_nparts", [1, 2, 5])
def test_pivot_with_col_names_subset(make_df, repartition_nparts, with_morsel_size):
    daft_df = make_df(
        {
            "group": ["A", "A", "B", "B"],
            "pivot": [1, 2, 1, 2],
            "value": [10, 20, 30, 40],
        },
        repartition=repartition_nparts,
    )
    daft_df = daft_df.pivot(
        group_by="group",
        pivot_col="pivot",
        value_col="value",
        agg_fn="sum",
        names=["1"],
    )

    expected = {
        "group": ["A", "B"],
        "1": [10, 30],
    }

    assert daft_df.sort("group").to_pydict() == expected


@pytest.mark.parametrize("repartition_nparts", [1, 2, 5])
def test_pivot_with_col_names_superset(make_df, repartition_nparts, with_morsel_size):
    daft_df = make_df(
        {
            "group": ["A", "A", "B", "B"],
            "pivot": [1, 2, 1, 2],
            "value": [10, 20, 30, 40],
        },
        repartition=repartition_nparts,
    )
    daft_df = daft_df.pivot(
        group_by="group",
        pivot_col="pivot",
        value_col="value",
        agg_fn="sum",
        names=["1", "2", "3"],
    )

    expected = {
        "group": ["A", "B"],
        "1": [10, 30],
        "2": [20, 40],
        "3": [None, None],
    }

    assert daft_df.sort("group").to_pydict() == expected


@pytest.mark.parametrize("repartition_nparts", [1, 2, 5])
def test_pivot_with_nulls(make_df, repartition_nparts, with_morsel_size):
    daft_df = make_df(
        {
            "group": ["A", None, "B", "B"],
            "pivot": [1, 2, None, 2],
            "value": [10, 20, 30, None],
        },
        repartition=repartition_nparts,
    )
    daft_df = daft_df.pivot(group_by="group", pivot_col="pivot", value_col="value", agg_fn="sum")

    expected = {
        "group": ["A", "B", None],
        "1": [10, None, None],
        "None": [None, 30, None],
        "2": [None, None, 20],
    }

    assert daft_df.sort("group").to_pydict() == expected


@pytest.mark.parametrize("repartition_nparts", [1, 2, 5])
@pytest.mark.parametrize(
    "agg_fn, expected",
    [
        ("sum", {"group": ["A", "B"], "1": [30, 70]}),
        ("mean", {"group": ["A", "B"], "1": [15, 35]}),
        ("max", {"group": ["A", "B"], "1": [20, 40]}),
        ("min", {"group": ["A", "B"], "1": [10, 30]}),
        ("count", {"group": ["A", "B"], "1": [2, 2]}),
    ],
)
def test_pivot_with_different_aggs(make_df, repartition_nparts, agg_fn, expected, with_morsel_size):
    daft_df = make_df(
        {
            "group": ["A", "A", "B", "B"],
            "pivot": [1, 1, 1, 1],
            "value": [10, 20, 30, 40],
        },
        repartition=repartition_nparts,
    )
    daft_df = daft_df.pivot(group_by="group", pivot_col="pivot", value_col="value", agg_fn=agg_fn)

    assert daft_df.sort("group").to_pydict() == expected


@pytest.mark.parametrize("repartition_nparts", [1, 2, 5])
def test_pivot_with_multiple_group_by(make_df, repartition_nparts, with_morsel_size):
    daft_df = make_df(
        {
            "group1": ["A", "A", "A", "B", "B", "B"],
            "group2": ["X", "X", "Y", "X", "Y", "Y"],
            "pivot": [1, 2, 1, 2, 1, 2],
            "value": [10, 20, 30, 40, 50, 60],
        },
        repartition=repartition_nparts,
    )
    daft_df = daft_df.pivot(group_by=["group1", "group2"], pivot_col="pivot", value_col="value", agg_fn="sum")

    expected = {
        "group1": ["A", "A", "B", "B"],
        "group2": ["X", "Y", "X", "Y"],
        "1": [10, 30, None, 50],
        "2": [20, None, 40, 60],
    }
    assert daft_df.sort(["group1", "group2"]).to_pydict() == expected


@pytest.mark.parametrize("repartition_nparts", [1, 2, 5])
def test_pivot_with_downstream_ops(make_df, repartition_nparts, with_morsel_size):
    daft_df = make_df(
        {
            "group": ["A", "A", "B", "B"],
            "pivot": [1, 2, 1, 2],
            "value": [10, 20, 30, 40],
        },
        repartition=repartition_nparts,
    )
    daft_df = daft_df.pivot(group_by="group", pivot_col="pivot", value_col="value", agg_fn="sum")
    daft_df = daft_df.where(daft_df["1"] == 10).select("2")

    assert daft_df.to_pydict() == {"2": [20]}



from pathlib import Path

import daft
from daft.sql.sql import SQLCatalog


def generate_catalog(dir: Path):
    if not dir.exists():
        raise RuntimeError(f"Directory not found: {dir}")
    return SQLCatalog(
        tables={
            file.stem: daft.read_parquet(path=str(file))
            for file in dir.iterdir()
            if file.is_file() and file.suffix == ".parquet"
        }
    )


def parse_questions_str(questions: str) -> list[int]:
    if questions == "*":
        return list(range(1, 100))

    nums = set()
    for split in filter(lambda str: str, questions.split(",")):
        try:
            num = int(split)
            nums.add(num)
        except ValueError:
            ints = split.split("-")
            assert (
                len(ints) == 2
            ), f"A range must include two numbers split by a dash (i.e., '-'); instead got '{split}'"
            [lower, upper] = ints
            try:
                lower = int(lower)
                upper = int(upper)
                assert lower <= upper
                for index in range(lower, upper + 1):
                    nums.add(index)
            except ValueError:
                raise ValueError(f"Invalid range: {split}")

    return nums



import pytest

from daft import col
from daft.datatype import DataType


@pytest.mark.parametrize("n_partitions", [2])
def test_unpivot(make_df, n_partitions, with_morsel_size):
    df = make_df(
        {
            "id": ["x", "y", "z"],
            "a": [1, 3, 5],
            "b": [2, 4, 6],
        },
        repartition=n_partitions,
    )

    df = df.unpivot("id", ["a", "b"])
    df = df.sort(["id", "variable"])
    df = df.collect()

    expected = {
        "id": ["x", "x", "y", "y", "z", "z"],
        "variable": ["a", "b", "a", "b", "a", "b"],
        "value": [1, 2, 3, 4, 5, 6],
    }

    assert df.to_pydict() == expected


@pytest.mark.parametrize("n_partitions", [1, 2, 4])
def test_unpivot_no_values(make_df, n_partitions, with_morsel_size):
    df = make_df(
        {
            "id": ["x", "y", "z"],
            "a": [1, 3, 5],
            "b": [2, 4, 6],
        },
        repartition=n_partitions,
    )

    df = df.unpivot("id")
    df = df.sort(["id", "variable"])
    df = df.collect()

    expected = {
        "id": ["x", "x", "y", "y", "z", "z"],
        "variable": ["a", "b", "a", "b", "a", "b"],
        "value": [1, 2, 3, 4, 5, 6],
    }

    assert df.to_pydict() == expected


@pytest.mark.parametrize("n_partitions", [1, 2, 4])
def test_unpivot_different_types(make_df, n_partitions, with_morsel_size):
    df = make_df(
        {
            "id": ["x", "y", "z"],
            "a": [1, 3, 5],
            "b": [1.1, 3.3, 5.5],
        },
        repartition=n_partitions,
    )

    df = df.unpivot("id", ["a", "b"])
    df = df.sort(["id", "variable"])
    df = df.collect()

    expected = {
        "id": ["x", "x", "y", "y", "z", "z"],
        "variable": ["a", "b", "a", "b", "a", "b"],
        "value": [1, 1.1, 3, 3.3, 5, 5.5],
    }

    assert df.to_pydict() == expected


@pytest.mark.parametrize("n_partitions", [1, 2, 4])
def test_unpivot_incompatible_types(make_df, n_partitions, with_morsel_size):
    df = make_df(
        {
            "id": ["x", "y", "z"],
            "a": [[1], [2, 3], [4, 5, 6]],
            "b": [7, 8, 9],
        },
        repartition=n_partitions,
    )

    with pytest.raises(ValueError):
        df = df.unpivot("id", ["a", "b"])


@pytest.mark.parametrize("n_partitions", [1, 2, 4])
def test_unpivot_nulls(make_df, n_partitions, with_morsel_size):
    df = make_df(
        {
            "id": ["x", "y", "z"],
            "a": [1, None, 5],
            "b": [2, 4, None],
        },
        repartition=n_partitions,
    )

    df = df.unpivot("id", ["a", "b"])
    df = df.sort(["id", "variable"])
    df = df.collect()

    expected = {
        "id": ["x", "x", "y", "y", "z", "z"],
        "variable": ["a", "b", "a", "b", "a", "b"],
        "value": [1, 2, None, 4, 5, None],
    }

    assert df.to_pydict() == expected


@pytest.mark.parametrize("n_partitions", [1, 2, 4])
def test_unpivot_null_column(make_df, n_partitions, with_morsel_size):
    df = make_df(
        {
            "id": ["x", "y", "z"],
            "a": [1, 3, 5],
            "b": [None, None, None],
        },
        repartition=n_partitions,
    )

    df = df.unpivot("id", ["a", "b"])
    df = df.sort(["id", "variable"])
    df = df.collect()

    expected = {
        "id": ["x", "x", "y", "y", "z", "z"],
        "variable": ["a", "b", "a", "b", "a", "b"],
        "value": [1, None, 3, None, 5, None],
    }

    assert df.to_pydict() == expected


@pytest.mark.parametrize("n_partitions", [1, 2, 4])
def test_unpivot_multiple_ids(make_df, n_partitions, with_morsel_size):
    df = make_df(
        {
            "id1": ["x", "y", "z"],
            "id2": [7, 8, 9],
            "a": [1, 3, 5],
            "b": [2, 4, 6],
        },
        repartition=n_partitions,
    )

    df = df.unpivot(["id1", "id2"], ["a", "b"])
    df = df.sort(["id1", "id2", "variable"])
    df = df.collect()

    expected = {
        "id1": ["x", "x", "y", "y", "z", "z"],
        "id2": [7, 7, 8, 8, 9, 9],
        "variable": ["a", "b", "a", "b", "a", "b"],
        "value": [1, 2, 3, 4, 5, 6],
    }

    assert df.to_pydict() == expected


@pytest.mark.parametrize("n_partitions", [1, 2, 4])
def test_unpivot_no_ids(make_df, n_partitions, with_morsel_size):
    df = make_df(
        {
            "a": [1, 3, 5],
            "b": [2, 4, 6],
        },
        repartition=n_partitions,
    )

    df = df.unpivot([])
    df = df.sort("value")
    df = df.collect()

    expected = {
        "variable": ["a", "b", "a", "b", "a", "b"],
        "value": [1, 2, 3, 4, 5, 6],
    }

    assert df.to_pydict() == expected


@pytest.mark.parametrize("n_partitions", [1, 2, 4])
def test_unpivot_expr(make_df, n_partitions, with_morsel_size):
    df = make_df(
        {
            "id": ["x", "y", "z"],
            "a": [1, 3, 5],
            "b": [2, 4, 6],
        },
        repartition=n_partitions,
    )

    df = df.unpivot("id", ["a", "b", (col("a") + col("b")).alias("a_plus_b")])
    df = df.sort(["id", "variable"])
    df = df.collect()

    expected = {
        "id": ["x", "x", "x", "y", "y", "y", "z", "z", "z"],
        "variable": ["a", "a_plus_b", "b", "a", "a_plus_b", "b", "a", "a_plus_b", "b"],
        "value": [1, 3, 2, 3, 7, 4, 5, 11, 6],
    }

    assert df.to_pydict() == expected


@pytest.mark.parametrize("n_partitions", [1, 2, 4])
def test_unpivot_followed_by_projection(make_df, n_partitions, with_morsel_size):
    df = make_df(
        {
            "year": [2020, 2021, 2022],
            "id": [1, 2, 3],
            "Jan": [10, 30, 50],
            "Feb": [20, 40, 60],
        },
        repartition=n_partitions,
    )

    df = df.unpivot("year", ["Jan", "Feb"], variable_name="month", value_name="inventory")
    df = df.with_column("inventory2", df["inventory"])
    df = df.sort(["year", "month", "inventory"])

    expected = {
        "year": [2020, 2020, 2021, 2021, 2022, 2022],
        "month": ["Feb", "Jan", "Feb", "Jan", "Feb", "Jan"],
        "inventory": [20, 10, 40, 30, 60, 50],
        "inventory2": [20, 10, 40, 30, 60, 50],
    }
    assert df.to_pydict() == expected


def test_unpivot_empty(make_df):
    df = make_df(
        {
            "id": [],
            "a": [],
            "b": [],
        }
    )

    df = df.unpivot("id", ["a", "b"])
    df = df.collect()

    expected = {
        "id": [],
        "variable": [],
        "value": [],
    }

    assert df.to_pydict() == expected
    assert df.schema()["variable"].dtype == DataType.string()


def test_unpivot_empty_partition(make_df):
    df = make_df(
        {
            "id": ["x", "y", "z"],
            "a": [1, 3, 5],
            "b": [2, 4, 6],
        }
    )

    df = df.into_partitions(4)
    df = df.unpivot("id", ["a", "b"])
    df = df.sort(["id", "variable"])
    df = df.collect()

    expected = {
        "id": ["x", "x", "y", "y", "z", "z"],
        "variable": ["a", "b", "a", "b", "a", "b"],
        "value": [1, 2, 3, 4, 5, 6],
    }

    assert df.to_pydict() == expected



from __future__ import annotations

from daft.daft import (
    AzureConfig,
    GCSConfig,
    IOConfig,
    HTTPConfig,
    S3Config,
    S3Credentials,
)
from daft.io._csv import read_csv
from daft.io._deltalake import read_deltalake
from daft.io._hudi import read_hudi
from daft.io._iceberg import read_iceberg
from daft.io._json import read_json
from daft.io._lance import read_lance
from daft.io._parquet import read_parquet
from daft.io._sql import read_sql
from daft.io.catalog import DataCatalogTable, DataCatalogType
from daft.io.file_path import from_glob_path

__all__ = [
    "AzureConfig",
    "DataCatalogTable",
    "DataCatalogType",
    "GCSConfig",
    "HTTPConfig",
    "IOConfig",
    "S3Config",
    "S3Credentials",
    "from_glob_path",
    "read_csv",
    "read_deltalake",
    "read_hudi",
    "read_iceberg",
    "read_json",
    "read_lance",
    "read_parquet",
    "read_sql",
]



from __future__ import annotations

from daft.expressions import col
from tests.expressions.typing.conftest import (
    assert_typing_resolve_vs_runtime_behavior,
    has_supertype,
)


def test_is_null(unary_data_fixture):
    arg = unary_data_fixture
    assert_typing_resolve_vs_runtime_behavior(
        data=(unary_data_fixture,),
        expr=col(arg.name()).is_null(),
        run_kernel=lambda: arg.is_null(),
        resolvable=True,
    )


def test_not_null(unary_data_fixture):
    arg = unary_data_fixture
    assert_typing_resolve_vs_runtime_behavior(
        data=(unary_data_fixture,),
        expr=col(arg.name()).not_null(),
        run_kernel=lambda: arg.not_null(),
        resolvable=True,
    )


def test_fill_null(binary_data_fixture):
    lhs, rhs = binary_data_fixture
    assert_typing_resolve_vs_runtime_behavior(
        data=binary_data_fixture,
        expr=col(lhs.name()).fill_null(col(rhs.name())),
        run_kernel=lambda: lhs.fill_null(rhs),
        resolvable=has_supertype(lhs.datatype(), rhs.datatype()),
    )



from __future__ import annotations

import shlex
import subprocess

import toml


def calculate_version() -> str:
    match_for_tag = r"v[0-9]*\.[0-9]*\.[0-9]*"
    exclude_for_tag = "*dev*"
    git_describe_base_cmd = f"git describe --match='{match_for_tag}' --exclude '{exclude_for_tag}' --tags"
    git_tag_version_number_command = f"{git_describe_base_cmd} --abbrev=0"

    git_tag_version_output = subprocess.check_output(shlex.split(git_tag_version_number_command))
    assert git_tag_version_output is not None
    cleaned_version = list(map(int, git_tag_version_output.decode().lstrip("v").rstrip("\n").split(".")))
    MAJOR, MINOR, PATCH = cleaned_version

    git_full_tag_version_output = subprocess.check_output(shlex.split(git_describe_base_cmd))
    assert git_full_tag_version_output is not None
    split_full_tag_version = git_full_tag_version_output.decode().strip().split("-")

    if len(split_full_tag_version) > 1:
        _, distance, commit = split_full_tag_version
        version = f"{MAJOR}.{MINOR}.{PATCH}+dev{int(distance):04d}.{commit[1:]}"
    else:
        version = f"{MAJOR}.{MINOR}.{PATCH}"
    return version


def patch_cargo_toml_version(version: str):
    CARGO_TOML = "Cargo.toml"
    source = toml.load(CARGO_TOML)

    # update package version
    assert "package" in source
    package = source["package"]
    assert "version" in package
    package["version"] = version

    # update workspace version
    assert "workspace" in source
    workspace = source["workspace"]
    assert "package" in workspace
    package = workspace["package"]
    assert "version" in package
    package["version"] = version

    with open("Cargo.toml", "w") as out_file:
        toml.dump(source, out_file)
    print(f"Patched {CARGO_TOML} to version: {version}")


if __name__ == "__main__":
    version = calculate_version()
    patch_cargo_toml_version(version=version)



from __future__ import annotations

import os
import uuid
from datetime import datetime

import pyarrow as pa
import pytest
from pyarrow import dataset as pads

import daft
from tests.conftest import assert_df_equals
from tests.cookbook.assets import COOKBOOK_DATA_CSV

PYARROW_GE_7_0_0 = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric()) >= (7, 0, 0)


def test_parquet_write(tmp_path, with_morsel_size):
    df = daft.read_csv(COOKBOOK_DATA_CSV)

    pd_df = df.write_parquet(tmp_path)
    read_back_pd_df = daft.read_parquet(tmp_path.as_posix() + "/*.parquet").to_pandas()
    assert_df_equals(df.to_pandas(), read_back_pd_df)

    assert len(pd_df) == 1
    assert pd_df._preview.preview_partition is None
    pd_df.__repr__()
    assert len(pd_df._preview.preview_partition) == 1


def test_parquet_write_with_partitioning(tmp_path, with_morsel_size):
    df = daft.read_csv(COOKBOOK_DATA_CSV)

    pd_df = df.write_parquet(tmp_path, partition_cols=["Borough"])

    read_back_pd_df = daft.read_parquet(tmp_path.as_posix() + "/**/*.parquet").to_pandas()
    assert_df_equals(df.to_pandas(), read_back_pd_df)

    assert len(pd_df) == 5
    assert pd_df._preview.preview_partition is None
    pd_df.__repr__()
    assert len(pd_df._preview.preview_partition) == 5


@pytest.mark.parametrize("write_mode", ["append", "overwrite"])
def test_empty_parquet_write_without_partitioning(tmp_path, write_mode, with_morsel_size):
    df = daft.read_csv(COOKBOOK_DATA_CSV)
    df = df.where(daft.lit(False))

    # Create a unique path to make sure that the writer is comfortable with nonexistent directories
    path = os.path.join(tmp_path, str(uuid.uuid4()))
    pd_df = df.write_parquet(path, write_mode=write_mode)
    read_back_pd_df = daft.read_parquet(path).to_pandas()
    assert_df_equals(df.to_pandas(), read_back_pd_df)

    assert len(pd_df) == 1
    assert len(pd_df._preview.preview_partition) == 1


@pytest.mark.parametrize("write_mode", ["append", "overwrite"])
def test_empty_parquet_write_with_partitioning(tmp_path, write_mode, with_morsel_size):
    df = daft.read_csv(COOKBOOK_DATA_CSV)
    df = df.where(daft.lit(False))

    # Create a unique path to make sure that the writer is comfortable with nonexistent directories
    path = os.path.join(tmp_path, str(uuid.uuid4()))
    output_files = df.write_parquet(path, partition_cols=["Borough"], write_mode=write_mode)
    read_back_pd_df = daft.read_parquet(os.path.join(path, "**/*.parquet")).to_pandas()
    assert_df_equals(df.to_pandas(), read_back_pd_df)

    assert len(output_files) == 1
    assert len(output_files._preview.preview_partition) == 1


def test_parquet_write_with_partitioning_readback_values(tmp_path, with_morsel_size):
    df = daft.read_csv(COOKBOOK_DATA_CSV)

    output_files = df.write_parquet(tmp_path, partition_cols=["Borough"])
    output_dict = output_files.to_pydict()
    boroughs = {"QUEENS", "MANHATTAN", "STATEN ISLAND", "BROOKLYN", "BRONX"}
    assert set(output_dict["Borough"]) == boroughs

    for path, bor in zip(output_dict["path"], output_dict["Borough"]):
        assert f"Borough={bor}" in path
        read_back = daft.read_parquet(path).to_pydict()
        assert all(b == bor for b in read_back["Borough"])

    read_back_pd_df = daft.read_parquet(tmp_path.as_posix() + "/**/*.parquet").to_pandas()
    assert_df_equals(df.to_pandas(), read_back_pd_df)

    assert len(output_files) == 5
    assert output_files._preview.preview_partition is None
    output_files.__repr__()
    assert len(output_files._preview.preview_partition) == 5


@pytest.mark.parametrize(
    "exp,key,answer",
    [
        (
            daft.col("date").partitioning.days(),
            "date_days",
            [19723, 19754, 19783, 19814, 19844],
        ),
        (daft.col("date").partitioning.hours(), "date_hours", [473352, 474096, 474792, 475536, 476256]),
        (daft.col("date").partitioning.months(), "date_months", [648, 649, 650, 651, 652]),
        (daft.col("date").partitioning.years(), "date_years", [54]),
    ],
)
def test_parquet_write_with_iceberg_date_partitioning(exp, key, answer, tmp_path, with_morsel_size):
    data = {
        "id": [1, 2, 3, 4, 5],
        "date": [
            datetime(2024, 1, 1),
            datetime(2024, 2, 1),
            datetime(2024, 3, 1),
            datetime(2024, 4, 1),
            datetime(2024, 5, 1),
        ],
    }
    df = daft.from_pydict(data)
    date_files = df.write_parquet(tmp_path, partition_cols=[exp]).sort(by=key)
    output_dict = date_files.to_pydict()
    assert len(output_dict[key]) == len(answer)
    assert output_dict[key] == answer
    read_back_pd_df = daft.read_parquet(tmp_path.as_posix() + "/**/*.parquet").to_pandas()
    assert_df_equals(df.to_pandas(), read_back_pd_df, sort_key="id")


@pytest.mark.parametrize(
    "exp,key,answer",
    [
        (daft.col("id").partitioning.iceberg_bucket(10), "id_bucket", [0, 3, 5, 6, 8]),
        (daft.col("id").partitioning.iceberg_truncate(10), "id_trunc", [0, 10, 20, 40]),
    ],
)
def test_parquet_write_with_iceberg_bucket_and_trunc(exp, key, answer, tmp_path, with_morsel_size):
    data = {
        "id": [1, 12, 23, 24, 45],
        "date": [
            datetime(2024, 1, 1),
            datetime(2024, 2, 1),
            datetime(2024, 3, 1),
            datetime(2024, 4, 1),
            datetime(2024, 5, 1),
        ],
    }
    df = daft.from_pydict(data)
    date_files = df.write_parquet(tmp_path, partition_cols=[exp]).sort(by=key)
    output_dict = date_files.to_pydict()
    assert len(output_dict[key]) == len(answer)
    assert output_dict[key] == answer
    read_back_pd_df = daft.read_parquet(tmp_path.as_posix() + "/**/*.parquet").to_pandas()
    assert_df_equals(df.to_pandas(), read_back_pd_df, sort_key="id")


@pytest.mark.skipif(
    not PYARROW_GE_7_0_0,
    reason="We only use pyarrow datasets 7 for this test",
)
def test_parquet_write_with_null_values(tmp_path, with_morsel_size):
    df = daft.from_pydict({"x": [1, 2, 3, None]})
    df.write_parquet(tmp_path, partition_cols=[df["x"].alias("y")])
    ds = pads.dataset(tmp_path, format="parquet", partitioning=pads.HivePartitioning(pa.schema([("y", pa.int64())])))
    readback = ds.to_table()
    assert readback.to_pydict() == {"x": [1, 2, 3, None], "y": [1, 2, 3, None]}


@pytest.fixture()
def smaller_parquet_target_filesize():
    with daft.execution_config_ctx(parquet_target_filesize=1024):
        yield


@pytest.mark.skipif(
    not PYARROW_GE_7_0_0,
    reason="We only use pyarrow datasets 7 for this test",
)
def test_parquet_write_multifile(tmp_path, smaller_parquet_target_filesize, with_morsel_size):
    data = {"x": list(range(1_000))}
    df = daft.from_pydict(data)
    df2 = df.write_parquet(tmp_path)
    assert len(df2) > 1
    read_back = daft.read_parquet(tmp_path.as_posix() + "/*.parquet").sort(by="x").to_pydict()
    assert read_back == data


@pytest.mark.skipif(
    not PYARROW_GE_7_0_0,
    reason="We only use pyarrow datasets 7 for this test",
)
def test_parquet_write_multifile_with_partitioning(tmp_path, smaller_parquet_target_filesize, with_morsel_size):
    data = {"x": list(range(1_000))}
    df = daft.from_pydict(data)
    df2 = df.write_parquet(tmp_path, partition_cols=[df["x"].alias("y") % 2])
    assert len(df2) >= 4
    ds = pads.dataset(tmp_path, format="parquet", partitioning=pads.HivePartitioning(pa.schema([("y", pa.int64())])))
    readback = ds.to_table()
    readback = readback.sort_by("x").to_pydict()
    assert readback["x"] == data["x"]
    assert readback["y"] == [y % 2 for y in data["x"]]


def test_parquet_write_with_some_empty_partitions(tmp_path, with_morsel_size):
    data = {"x": [1, 2, 3], "y": ["a", "b", "c"]}
    daft.from_pydict(data).into_partitions(4).write_parquet(tmp_path)

    read_back = daft.read_parquet(tmp_path.as_posix() + "/**/*.parquet").sort("x").to_pydict()
    assert read_back == data


def test_parquet_partitioned_write_with_some_empty_partitions(tmp_path, with_morsel_size):
    data = {"x": [1, 2, 3], "y": ["a", "b", "c"]}
    output_files = daft.from_pydict(data).into_partitions(4).write_parquet(tmp_path, partition_cols=["x"])

    assert len(output_files) == 3

    read_back = daft.read_parquet(tmp_path.as_posix() + "/**/*.parquet").sort("x").to_pydict()
    assert read_back == data


def test_csv_write(tmp_path, with_morsel_size):
    df = daft.read_csv(COOKBOOK_DATA_CSV)

    pd_df = df.write_csv(tmp_path)

    read_back_pd_df = daft.read_csv(tmp_path.as_posix() + "/*.csv").to_pandas()
    assert_df_equals(df.to_pandas(), read_back_pd_df)

    assert len(pd_df) == 1
    assert pd_df._preview.preview_partition is None
    pd_df.__repr__()
    assert len(pd_df._preview.preview_partition) == 1


def test_csv_write_with_partitioning(tmp_path, with_morsel_size):
    df = daft.read_csv(COOKBOOK_DATA_CSV)
    schema = df.schema()
    names = schema.column_names()
    types = {}
    for n in names:
        types[n] = schema[n].dtype

    pd_df = df.write_csv(tmp_path, partition_cols=["Borough"]).to_pandas()
    read_back_pd_df = daft.read_csv(tmp_path.as_posix() + "/**/*.csv", schema=types).to_pandas()
    assert_df_equals(df.to_pandas().fillna(""), read_back_pd_df.fillna(""))

    assert len(pd_df) == 5


def test_empty_csv_write(tmp_path, with_morsel_size):
    df = daft.read_csv(COOKBOOK_DATA_CSV)
    df = df.where(daft.lit(False))

    schema = df.schema()
    names = schema.column_names()
    types = {}
    for n in names:
        types[n] = schema[n].dtype

    pd_df = df.write_csv(tmp_path)
    read_back_pd_df = daft.read_csv(tmp_path.as_posix() + "/*.csv", schema=types).to_pandas()
    assert_df_equals(df.to_pandas().fillna(""), read_back_pd_df.fillna(""))

    assert len(pd_df) == 1
    assert len(pd_df._preview.preview_partition) == 1


def test_empty_csv_write_with_partitioning(tmp_path, with_morsel_size):
    df = daft.read_csv(COOKBOOK_DATA_CSV)
    df = df.where(daft.lit(False))

    schema = df.schema()
    names = schema.column_names()
    types = {}
    for n in names:
        types[n] = schema[n].dtype

    pd_df = df.write_csv(tmp_path, partition_cols=["Borough"])
    read_back_pd_df = daft.read_csv(tmp_path.as_posix() + "/**/*.csv", schema=types).to_pandas()
    assert_df_equals(df.to_pandas().fillna(""), read_back_pd_df.fillna(""))

    assert len(pd_df) == 1
    assert len(pd_df._preview.preview_partition) == 1


def test_csv_write_with_some_empty_partitions(tmp_path, with_morsel_size):
    data = {"x": [1, 2, 3], "y": ["a", "b", "c"]}
    daft.from_pydict(data).into_partitions(4).write_csv(tmp_path)

    read_back = daft.read_csv(tmp_path.as_posix() + "/**/*.csv").sort("x").to_pydict()
    assert read_back == data


def test_csv_partitioned_write_with_some_empty_partitions(tmp_path, with_morsel_size):
    data = {"x": [1, 2, 3], "y": ["a", "b", "c"]}
    output_files = daft.from_pydict(data).into_partitions(4).write_csv(tmp_path, partition_cols=["x"])

    assert len(output_files) == 3

    read_back = daft.read_csv(tmp_path.as_posix() + "/**/*.csv").sort("x").to_pydict()
    assert read_back == data



from __future__ import annotations

import sys

from daft.dependencies import pa


def ensure_array(arr: pa.Array) -> pa.Array:
    """Applies all fixes to an Arrow array."""
    arr = _FixEmptyStructArrays.ensure_array(arr)
    arr = _FixSliceOffsets.ensure_array(arr)
    return arr


def ensure_chunked_array(arr: pa.ChunkedArray) -> pa.ChunkedArray:
    """Applies all fixes to an Arrow chunked array."""
    arr = _FixEmptyStructArrays.ensure_chunked_array(arr)
    arr = _FixSliceOffsets.ensure_chunked_array(arr)
    return arr


def ensure_table(tbl: pa.Table) -> pa.Table:
    """Applies all fixes to an Arrow table."""
    tbl = _FixEmptyStructArrays.ensure_table(tbl)
    tbl = _FixSliceOffsets.ensure_table(tbl)
    return tbl


class _FixEmptyStructArrays:
    """Converts StructArrays that are empty (have no fields) to StructArrays with a single field named "" and with a NullType.

    This is done because arrow2::ffi cannot handle empty StructArrays and we need to handle this on the
    Python layer before going through ffi into Rust.
    """

    @staticmethod
    def get_empty_struct_type():
        return pa.struct([])

    @staticmethod
    def get_single_field_struct_type():
        return pa.struct({"": pa.null()})

    @staticmethod
    def get_single_field_struct_value():
        return {"": None}

    def ensure_table(table: pa.Table) -> pa.Table:
        empty_struct_fields = [
            (i, f) for (i, f) in enumerate(table.schema) if f.type == _FixEmptyStructArrays.get_empty_struct_type()
        ]
        if not empty_struct_fields:
            return table
        for i, field in empty_struct_fields:
            table = table.set_column(i, field.name, _FixEmptyStructArrays.ensure_chunked_array(table[field.name]))
        return table

    def ensure_chunked_array(arr: pa.ChunkedArray) -> pa.ChunkedArray:
        if arr.type != _FixEmptyStructArrays.get_empty_struct_type():
            return arr
        return pa.chunked_array([_FixEmptyStructArrays.ensure_array(chunk) for chunk in arr.chunks])

    def ensure_array(arr: pa.Array) -> pa.Array:
        """Recursively converts empty struct arrays to single-field struct arrays."""
        if arr.type == _FixEmptyStructArrays.get_empty_struct_type():
            return pa.array(
                [
                    _FixEmptyStructArrays.get_single_field_struct_value() if valid.as_py() else None
                    for valid in arr.is_valid()
                ],
                type=_FixEmptyStructArrays.get_single_field_struct_type(),
            )

        elif isinstance(arr, pa.StructArray):
            new_arrays = [ensure_array(arr.field(field.name)) for field in arr.type]
            names = [field.name for field in arr.type]
            null_mask = arr.is_null()

            return pa.StructArray.from_arrays(new_arrays, names=names, mask=null_mask)

        else:
            return arr


def remove_empty_struct_placeholders(arr: pa.Array):
    """Recursively removes the empty struct placeholders placed by _FixEmptyStructArrays.ensure_array."""
    if arr.type == _FixEmptyStructArrays.get_single_field_struct_type():
        return pa.array(
            [{} if valid.as_py() else None for valid in arr.is_valid()],
            type=_FixEmptyStructArrays.get_empty_struct_type(),
        )

    elif isinstance(arr, pa.StructArray):
        new_arrays = [remove_empty_struct_placeholders(arr.field(field.name)) for field in arr.type]
        names = [field.name for field in arr.type]
        null_mask = arr.is_null()

        return pa.StructArray.from_arrays(new_arrays, names=names, mask=null_mask)

    else:
        return arr


class _FixSliceOffsets:
    # TODO(Clark): For pyarrow < 12.0.0, struct array slice offsets are dropped
    # when converting to record batches. We work around this below by flattening
    # the field arrays for all struct arrays, which propagates said offsets to
    # the child (field) arrays. Note that this flattening is zero-copy unless
    # the column contains a validity bitmap (i.e. the column has one or more nulls).
    #
    # We should enforce a lower Arrow version bound once the upstream fix is
    # released.
    #
    # See:
    #  - https://github.com/apache/arrow/issues/34639
    #  - https://github.com/apache/arrow/pull/34691

    @staticmethod
    def ensure_table(arrow_table: pa.Table) -> pa.Table:
        """Ensures that table-level slice offsets are properly propagated to child arrays to prevent them from being dropped upon record batch conversion and FFI transfer."""
        arrow_schema = arrow_table.schema
        for idx, name in enumerate(arrow_schema.names):
            field = arrow_schema.field(name)
            column = arrow_table.column(name)
            if _FixSliceOffsets._chunked_array_needs_slice_offset_propagation(column):
                new_column = _FixSliceOffsets._propagate_chunked_array_slice_offsets(column)
                arrow_table = arrow_table.set_column(idx, field, new_column)
        return arrow_table

    @staticmethod
    def ensure_chunked_array(chunked_array: pa.ChunkedArray) -> pa.ChunkedArray:
        """Ensures that chunked-array-level slice offsets are properly propagated to child arrays to prevent them from being dropped upon record batch conversion and FFI transfer."""
        if _FixSliceOffsets._chunked_array_needs_slice_offset_propagation(chunked_array):
            return _FixSliceOffsets._propagate_chunked_array_slice_offsets(chunked_array)
        else:
            return chunked_array

    @staticmethod
    def ensure_array(array: pa.Array) -> pa.Array:
        """Ensures that array-level slice offsets are properly propagated to child arrays to prevent them from being dropped upon record batch conversion and FFI transfer."""
        if _FixSliceOffsets._array_needs_slice_offset_propagation(array):
            return _FixSliceOffsets._propagate_array_slice_offsets(array)
        else:
            return array

    @staticmethod
    def _chunked_array_needs_slice_offset_propagation(chunked_array: pa.ChunkedArray) -> bool:
        """Whether an Arrow ChunkedArray needs slice offset propagation.

        This is currently only true for struct arrays and fixed-size list arrays that contain
        slice offsets/truncations.
        """
        if not pa.types.is_struct(chunked_array.type) and not pa.types.is_fixed_size_list(chunked_array.type):
            return False
        return any(_FixSliceOffsets._array_needs_slice_offset_propagation(chunk) for chunk in chunked_array.chunks)

    @staticmethod
    def _array_needs_slice_offset_propagation(array: pa.Array) -> bool:
        """Whether an Arrow array needs slice offset propagation.

        This is currently only true for struct arrays and fixed-size list arrays that contain
        slice offsets/truncations.
        """
        if pa.types.is_struct(array.type):
            return _FixSliceOffsets._struct_array_needs_slice_offset_propagation(array)
        elif pa.types.is_fixed_size_list(array.type):
            return _FixSliceOffsets._fixed_size_list_array_needs_slice_offset_propagation(array)
        else:
            return False

    @staticmethod
    def _struct_array_needs_slice_offset_propagation(array: pa.StructArray) -> bool:
        """Whether the provided struct array needs slice offset propagation."""
        assert isinstance(array, pa.StructArray)
        # TODO(Clark): Only propagate slice offsets if a slice exists; checking whether the
        # array length has been truncated is currently difficult since StructArray.field()
        # propagates the slice to the field arrays, and there's no other convenient way to
        # access the child field arrays other than reconstructing them from the raw buffers,
        # which is complex.
        # if array.type.num_fields == 0:
        #     return False
        # return array.offset > 0 or len(array) < len(array.field(0))
        return True

    @staticmethod
    def _fixed_size_list_array_needs_slice_offset_propagation(array: pa.FixedSizeListArray) -> bool:
        """Whether the provided fixed-size list array needs slice offset propagation."""
        assert isinstance(array, pa.FixedSizeListArray)
        return array.offset > 0 or len(array) < array.type.list_size * len(array.values)

    @staticmethod
    def _propagate_chunked_array_slice_offsets(chunked_array: pa.ChunkedArray) -> pa.ChunkedArray:
        """Propagate slice offsets for the provided chunked array to the child arrays of each chunk."""
        new_chunks = []
        # Flatten each chunk to propagate slice offsets to child arrays.
        for chunk in chunked_array.chunks:
            new_chunk = _FixSliceOffsets._propagate_array_slice_offsets(chunk)
            new_chunks.append(new_chunk)
        return pa.chunked_array(new_chunks)

    @staticmethod
    def _propagate_array_slice_offsets(array: pa.Array) -> pa.Array:
        """Propagate slice offsets for the provided array to its child arrays."""
        assert _FixSliceOffsets._array_needs_slice_offset_propagation(array)
        dtype = array.type
        if pa.types.is_struct(dtype):
            fields = [dtype[i] for i in range(dtype.num_fields)]
            # StructArray.flatten() will propagate slice offsets to the underlying field arrays
            # while also propagating the StructArray-level bitmap.
            new_field_arrays = array.flatten()
            return pa.StructArray.from_arrays(new_field_arrays, fields=fields, mask=array.is_null())
        elif pa.types.is_fixed_size_list(dtype):
            # TODO(Clark): FixedSizeListArray.flatten() doesn't properly handle bitmap propagation
            # to the child array, instead returning a concatenation of non-null array fragments.
            # We therefore manually slice the full values array and take care of the bitmap
            # propagation ourselves.
            # new_array = array.flatten()
            new_array = array.values.slice(array.offset * dtype.list_size, len(array) * dtype.list_size)
            bitmap_buf = array.buffers()[0]
            if bitmap_buf is not None:
                bitmap_buf = _FixSliceOffsets._slice_bitmap_buffer(bitmap_buf, array.offset, len(array))
            out = pa.FixedSizeListArray.from_buffers(dtype, len(array), [bitmap_buf], offset=0, children=[new_array])
            return out
        else:
            raise ValueError(f"Array type doesn't need array slice offset propagation: {dtype}")

    @staticmethod
    def _slice_bitmap_buffer(buf: pa.Buffer, offset: int, length: int) -> pa.Buffer:
        """Slice the provided bitpacked boolean bitmap buffer at the given offset and length.

        This function takes care of the byte and bit offset bookkeeping required due to the buffer
        being bitpacked.
        """
        # Offset to the leftmost byte of the buffer slice.
        byte_offset = offset // 8
        # Bit offset into the leftmost byte of the buffer slice.
        bit_offset = offset % 8
        # The byte-padded number of bits for the buffer slice.
        # This rounds up the buffer slice to the nearest multiple of 8.
        num_bytes_for_bits = (bit_offset + length + 7) & (-8)
        # The number of bytes in the byte-padded buffer slice.
        byte_length = num_bytes_for_bits // 8
        buf = buf.slice(byte_offset, byte_length)
        if bit_offset != 0:
            # If we have a bit offset into the leftmost byte of the buffer slice,
            # we need to bitshift the buffer to eliminate the bit offset.
            bytes_ = buf.to_pybytes()
            bytes_as_int = int.from_bytes(bytes_, sys.byteorder)
            bytes_as_int >>= bit_offset
            bytes_ = bytes_as_int.to_bytes(byte_length, sys.byteorder)
            buf = pa.py_buffer(bytes_)
        return buf



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from daft.dataframe.preview import DataFramePreview
    from daft.logical.schema import Schema


@dataclass(frozen=True)
class DataFrameDisplay:
    preview: DataFramePreview
    schema: Schema
    # These formatting options are deprecated for now and not guaranteed to be supported.
    column_char_width: int = 20
    max_col_rows: int = 3
    num_rows: int = 10

    def _get_user_message(self) -> str:
        if self.preview.preview_partition is None:
            return "(No data to display: Dataframe not materialized)"
        if self.preview.dataframe_num_rows == 0:
            return "(No data to display: Materialized dataframe has no rows)"
        if self.preview.dataframe_num_rows is None:
            return f"(Showing first {min(self.num_rows, len(self.preview.preview_partition))} rows)"
        return f"(Showing first {min(self.num_rows, len(self.preview.preview_partition))} of {self.preview.dataframe_num_rows} rows)"

    def _repr_html_(self) -> str:
        if len(self.schema) == 0:
            return "<small>(No data to display: Dataframe has no columns)</small>"

        res = "<div>\n"

        if self.preview.preview_partition is not None:
            res += self.preview.preview_partition.to_table()._repr_html_()
        else:
            res += self.schema._truncated_table_html()

        res += f"\n<small>{self._get_user_message()}</small>\n</div>"

        return res

    def __repr__(self) -> str:
        if len(self.schema) == 0:
            return "(No data to display: Dataframe has no columns)"

        if self.preview.preview_partition is not None:
            res = repr(self.preview.preview_partition.to_table())
        else:
            res = self.schema._truncated_table_string()

        res += f"\n{self._get_user_message()}"

        return res



from __future__ import annotations

import logging
import os
from contextlib import contextmanager
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from viztracer import VizTracer

ACTIVE = False


logger = logging.getLogger(__name__)


@contextmanager
def profiler(filename: str) -> VizTracer:
    if int(os.environ.get("DAFT_PROFILING", 0)) == 1:
        from viztracer import VizTracer, get_tracer

        global ACTIVE
        if not ACTIVE:
            tracer = get_tracer()
            if tracer is None or tracer.output_file != filename:
                tracer = VizTracer(output_file=filename, tracer_entries=10_000_000)
            ACTIVE = True
            with tracer:
                yield tracer
            ACTIVE = False
            return
        else:
            tracer = get_tracer()
            logger.warning(
                "profiler(%s) not created. Another profiler(%s) is already active.", filename, tracer.output_file
            )

    yield None
    return


import time


@contextmanager
def timingcontext(name: str):
    from viztracer import get_tracer

    tracer = get_tracer()

    try:
        logger.debug("log_event:enter:%s", name)
        start = time.time()
        if tracer is not None:
            with tracer.log_event(name) as event:
                yield event
        else:
            yield None
    finally:
        end = time.time()

        logger.debug("log_event:%s:%sms", name, f"{(end-start)*1000:.3f}")


def log_event(name: str):
    return timingcontext(name)



from __future__ import annotations

import boto3

s3 = boto3.client("s3")

all_objects = []

paginator = s3.get_paginator("list_objects_v2")
pages = paginator.paginate(Bucket="github-actions-artifacts-bucket")

for page in pages:
    for obj in page["Contents"]:
        all_objects.append(obj)

all_wheels = [obj["Key"] for obj in all_objects if obj["Key"].endswith(".whl")]

html_href = []

for key in all_wheels:
    name = key.split("/")[-1]
    link = key.replace("+", "%2B")
    href = f"    <a href='../{link}'>{name} </a>"
    html_href.append(href)
    html_href.append("    <br />")

html_href.pop()

all_href = "\n".join(h for h in html_href)

manifest = f"""<html>
<head>
    <title>Links</title>
</head>
<body>
    <h1>Links</h1>
    {all_href}
</body>
</html>"""


with open("index.html", "w") as f:
    f.write(manifest)

s3.upload_file(
    "index.html",
    Bucket="github-actions-artifacts-bucket",
    Key="getdaft/index.html",
    ExtraArgs={"ACL": "public-read", "ContentType": "text/html"},
)

print("uploaded:")
print(manifest)



from __future__ import annotations

import pytest

import daft
from daft import DataFrame
from daft.recordbatch import MicroPartition

NUM_ROWS = 1_000_000


# Perform filter against a int64 column: selecting every other element
def generate_int64_keep_every_other() -> tuple[dict, daft.Expression, list]:
    return (
        {"data": list(range(NUM_ROWS)), "mask": [True if i % 2 == 0 else False for i in range(NUM_ROWS)]},
        [i for i in range(NUM_ROWS) if i % 2 == 0],
    )


# Perform filter against a int64 column: selecting every third element, where every two elements in between are False and Null
def generate_int64_keep_every_other_nulls() -> tuple[dict, daft.Expression, list]:
    return (
        {
            "data": list(range(NUM_ROWS)),
            "mask": [True if i % 3 == 0 else (False if i % 3 == 1 else None) for i in range(NUM_ROWS)],
        },
        [i for i in range(NUM_ROWS) if i % 3 == 0],
    )


# Perform filter against a int64 column: contiguous segments of True/False
def generate_int64_keep_contiguous_chunk() -> tuple[dict, daft.Expression, list]:
    return (
        {
            "data": list(range(NUM_ROWS)),
            "mask": [True for _ in range(NUM_ROWS // 2)] + [False for _ in range(NUM_ROWS // 2)],
        },
        list(range(NUM_ROWS // 2)),
    )


# Perform filter against a list[int64] column: selecting every other element
def generate_list_int64_keep_every_other() -> tuple[dict, daft.Expression, list]:
    data = [list(range(i, i + 4)) for i in range(NUM_ROWS)]
    return (
        {"data": data, "mask": [True if i % 2 == 0 else False for i in range(NUM_ROWS)]},
        [d for i, d in enumerate(data) if i % 2 == 0],
    )


# Perform filter against a list[int64] column: selecting every third element, where every two elements in between are False and Null
def generate_list_int64_keep_every_other_nulls() -> tuple[dict, daft.Expression, list]:
    data = [list(range(i, i + 4)) for i in range(NUM_ROWS)]
    return (
        {"data": data, "mask": [True if i % 3 == 0 else (False if i % 3 == 1 else None) for i in range(NUM_ROWS)]},
        [d for i, d in enumerate(data) if i % 3 == 0],
    )


# Perform filter against a list[int64] column: contiguous segments of True/False
def generate_list_int64_keep_contiguous_chunks() -> tuple[dict, daft.Expression, list]:
    data = [list(range(i, i + 4)) for i in range(NUM_ROWS)]
    return (
        {"data": data, "mask": [int(i // 100) % 2 == 0 for i in range(NUM_ROWS)]},
        [d for i, d in enumerate(data) if int(i // 100) % 2 == 0],
    )


# Perform filter against a list[int64] column: keep all values
def generate_list_int64_keep_all() -> tuple[dict, daft.Expression, list]:
    data = [list(range(i, i + 4)) for i in range(NUM_ROWS)]
    return (
        {"data": data, "mask": [True for _ in range(NUM_ROWS)]},
        data,
    )


# Perform filter against a list[int64] column: keep none of the values
def generate_list_int64_keep_none() -> tuple[dict, daft.Expression, list]:
    data = [list(range(i, i + 4)) for i in range(NUM_ROWS)]
    return (
        {"data": data, "mask": [False for _ in range(NUM_ROWS)]},
        [],
    )


@pytest.mark.benchmark(group="if_else")
@pytest.mark.parametrize(
    "test_data_generator",
    [
        pytest.param(
            generate_int64_keep_every_other,
            id="int64-every-other",
        ),
        pytest.param(
            generate_int64_keep_every_other_nulls,
            id="int64-every-other-with-nulls",
        ),
        pytest.param(
            generate_int64_keep_contiguous_chunk,
            id="int64-contiguous",
        ),
        pytest.param(
            generate_list_int64_keep_every_other,
            id="int64-list-every-other",
        ),
        pytest.param(
            generate_list_int64_keep_every_other_nulls,
            id="int64-list-every-other-with-nulls",
        ),
        pytest.param(
            generate_list_int64_keep_contiguous_chunks,
            id="int64-list-contiguous",
        ),
        pytest.param(
            generate_list_int64_keep_all,
            id="int64-list-all",
        ),
        pytest.param(
            generate_list_int64_keep_none,
            id="int64-list-none",
        ),
    ],
)
def test_filter(test_data_generator, benchmark) -> None:
    """If_else between NUM_ROWS values."""
    data, expected = test_data_generator()
    table = MicroPartition.from_pydict(data)

    def bench_filter() -> DataFrame:
        return table.filter([daft.col("mask")])

    result = benchmark(bench_filter)
    assert result.to_pydict()["data"] == expected



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass

from daft.dependencies import pa, pafs
from daft.filesystem import join_path
from daft.hudi.pyhudi.filegroup import BaseFile, FileGroup, FileSlice
from daft.hudi.pyhudi.timeline import Timeline
from daft.hudi.pyhudi.utils import (
    list_full_sub_dirs,
    list_leaf_dirs,
    list_relative_file_paths,
)

# TODO(Shiyan): support base file in .orc
BASE_FILE_EXTENSIONS = [".parquet"]

HUDI_METAFIELD_PARTITION_PATH = "_hoodie_partition_path"


@dataclass
class MetaClient:
    fs: pafs.FileSystem
    base_path: str
    timeline: Timeline | None

    def get_active_timeline(self) -> Timeline:
        if not self.timeline:
            self.timeline = Timeline(self.base_path, self.fs)
        return self.timeline

    def get_partition_paths(self, relative=True) -> list[str]:
        first_level_full_partition_paths = list_full_sub_dirs(self.base_path, self.fs, excludes=[".hoodie"])
        partition_paths = []
        for p in first_level_full_partition_paths:
            partition_paths.extend(list_leaf_dirs(p, self.fs))

        common_prefix_len = len(self.base_path) + 1 if relative else 0
        return [p[common_prefix_len:] for p in partition_paths]

    def get_file_groups(self, partition_path: str) -> list[FileGroup]:
        base_file_metadata = list_relative_file_paths(
            self.base_path, partition_path, self.fs, includes=BASE_FILE_EXTENSIONS
        )
        fg_id_to_base_files = defaultdict(list)
        for metadata in base_file_metadata:
            base_file = BaseFile(metadata)
            fg_id_to_base_files[base_file.file_group_id].append(base_file)
        file_groups = []
        for fg_id, base_files in fg_id_to_base_files.items():
            file_group = FileGroup(fg_id, partition_path)
            for base_file in base_files:
                file_group.add_base_file(base_file)
            file_groups.append(file_group)
        return file_groups


@dataclass(init=False)
class FileSystemView:
    def __init__(self, meta_client: MetaClient):
        self.meta_client = meta_client
        self.partition_to_file_groups: dict[str, list[FileGroup]] = {}
        self._load_partitions()

    def _load_partitions(self):
        partition_paths = self.meta_client.get_partition_paths()
        if not partition_paths:
            self._load_partition("")
        else:
            for partition_path in partition_paths:
                self._load_partition(partition_path)

    def _load_partition(self, partition_path: str):
        file_groups = self.meta_client.get_file_groups(partition_path)
        self.partition_to_file_groups[partition_path] = file_groups

    def get_latest_file_slices(self) -> list[FileSlice]:
        file_slices = []
        for file_groups in self.partition_to_file_groups.values():
            for file_group in file_groups:
                file_slice = file_group.get_latest_file_slice()
                if file_slice is not None:
                    file_slices.append(file_slice)

        return file_slices


@dataclass(init=False)
class HudiTableProps:
    def __init__(self, fs: pafs.FileSystem, base_path: str):
        self._props = {}
        hoodie_properties_file = join_path(fs, base_path, ".hoodie", "hoodie.properties")
        with fs.open_input_file(hoodie_properties_file) as f:
            lines = f.readall().decode("utf-8").splitlines()
            for line in lines:
                line = line.strip()
                if not line or line.startswith("#"):
                    continue
                key, value = line.split("=")
                self._props[key] = value
        self._validate()
        self.is_partitioned = not self.keygen_classname.endswith("NonpartitionedKeyGenerator")

    @property
    def name(self) -> str:
        return self._props["hoodie.table.name"]

    @property
    def partition_fields(self) -> list[str]:
        if "hoodie.table.partition.fields" not in self._props:
            return []

        return self._props["hoodie.table.partition.fields"].split(",")

    @property
    def keygen_classname(self) -> str:
        return self._props["hoodie.table.keygenerator.class"]

    @property
    def is_hive_style_partitioning(self) -> bool:
        return bool(self._props["hoodie.datasource.write.hive_style_partitioning"])

    @property
    def populates_meta_fields(self) -> bool:
        return bool(self._props.get("hoodie.populate.meta.fields", "true"))

    def get_required_config(self, key: str) -> str:
        if key not in self._props:
            raise UnsupportedException(f"Missing required table config: {key}")
        return self._props[key]

    def _validate(self):
        if self.get_required_config("hoodie.table.type") != "COPY_ON_WRITE":
            raise UnsupportedException("Only support COPY_ON_WRITE table")
        if self.get_required_config("hoodie.table.version") not in ["5", "6"]:
            raise UnsupportedException("Only support table version 5 and 6")
        if self.get_required_config("hoodie.timeline.layout.version") != "1":
            raise UnsupportedException("Only support timeline layout version 1")
        if self.get_required_config("hoodie.datasource.write.drop.partition.columns") == "true":
            raise UnsupportedException("Only support when `hoodie.datasource.write.drop.partition.columns` is disabled")


@dataclass
class HudiTableMetadata:
    files_metadata: pa.RecordBatch
    colstats_min_values: pa.RecordBatch
    colstats_max_values: pa.RecordBatch


class UnsupportedException(Exception):
    pass


@dataclass(init=False)
class HudiTable:
    def __init__(self, table_uri: str, fs: pafs.FileSystem, resolved_base_path: str):
        self.table_uri = HudiTable._sanitize_path(table_uri)
        base_path = HudiTable._sanitize_path(resolved_base_path)
        self._meta_client = MetaClient(fs, base_path, timeline=None)
        self._props = HudiTableProps(fs, base_path)

    @staticmethod
    def _sanitize_path(path: str) -> str:
        return path.rstrip("/")

    def latest_table_metadata(self) -> HudiTableMetadata:
        file_slices = FileSystemView(self._meta_client).get_latest_file_slices()
        files_metadata = []
        min_vals_arr = []
        max_vals_arr = []
        files_metadata_schema = FileSlice.FILES_METADATA_SCHEMA if file_slices else pa.schema([])
        colstats_schema = file_slices[0].colstats_schema if file_slices else pa.schema([])
        for file_slice in file_slices:
            files_metadata.append(file_slice.files_metadata)
            min_vals, max_vals = file_slice.colstats_min_max
            min_vals_arr.append(min_vals)
            max_vals_arr.append(max_vals)
        metadata_arrays = [pa.array(column) for column in list(zip(*files_metadata))]
        min_value_arrays = [pa.array(column) for column in list(zip(*min_vals_arr))]
        max_value_arrays = [pa.array(column) for column in list(zip(*max_vals_arr))]
        return HudiTableMetadata(
            pa.RecordBatch.from_arrays(metadata_arrays, schema=files_metadata_schema),
            pa.RecordBatch.from_arrays(min_value_arrays, schema=colstats_schema),
            pa.RecordBatch.from_arrays(max_value_arrays, schema=colstats_schema),
        )

    @property
    def base_path(self) -> str:
        return self._meta_client.base_path

    @property
    def schema(self) -> pa.Schema:
        return self._meta_client.get_active_timeline().get_latest_commit_schema()

    @property
    def is_partitioned(self) -> bool:
        return self._props.is_partitioned

    @property
    def supports_partition_values(self) -> bool:
        return self._props.is_partitioned and self._props.populates_meta_fields

    @property
    def props(self) -> HudiTableProps:
        return self._props



from __future__ import annotations

import datetime

import pytest

pyiceberg = pytest.importorskip("pyiceberg")

import pyarrow.compute as pc
from pyiceberg.io.pyarrow import schema_to_pyarrow

import daft
from daft.logical.schema import Schema
from tests.conftest import assert_df_equals


@pytest.mark.integration()
def test_daft_iceberg_table_open(local_iceberg_tables, local_iceberg_catalog):
    catalog_name, pyiceberg_catalog = local_iceberg_catalog
    tab = pyiceberg_catalog.load_table(local_iceberg_tables)
    df = daft.read_table(f"{catalog_name}.{local_iceberg_tables}")
    iceberg_schema = tab.schema()
    as_pyarrow_schema = schema_to_pyarrow(iceberg_schema)
    as_daft_schema = Schema.from_pyarrow_schema(as_pyarrow_schema)
    assert df.schema() == as_daft_schema


WORKING_SHOW_COLLECT = [
    # "test_all_types",  # Commented out due to issue https://github.com/Eventual-Inc/Daft/issues/2996
    "test_limit",
    "test_null_nan",
    "test_null_nan_rewritten",
    "test_partitioned_by_bucket",
    "test_partitioned_by_days",
    "test_partitioned_by_hours",
    "test_partitioned_by_identity",
    "test_partitioned_by_months",
    "test_partitioned_by_truncate",
    "test_partitioned_by_years",
    "test_positional_mor_deletes",
    "test_positional_mor_double_deletes",
    # "test_table_sanitized_character", # Bug in scan().to_arrow().to_arrow()
    "test_table_version",  # we have bugs when loading no files
    "test_uuid_and_fixed_unpartitioned",
    "test_add_new_column",
    "test_new_column_with_no_data",
    "test_table_rename",
    # Partition evolution currently not supported, see issue: https://github.com/Eventual-Inc/Daft/issues/2249
    # "test_evolve_partitioning",
]


@pytest.mark.integration()
@pytest.mark.parametrize("table_name", WORKING_SHOW_COLLECT)
def test_daft_iceberg_table_show(table_name, local_iceberg_catalog):
    catalog_name, _ = local_iceberg_catalog
    df = daft.read_table(f"{catalog_name}.default.{table_name}")
    df.show()


@pytest.mark.integration()
@pytest.mark.parametrize("table_name", WORKING_SHOW_COLLECT)
def test_daft_iceberg_table_collect_correct(table_name, local_iceberg_catalog):
    catalog_name, pyiceberg_catalog = local_iceberg_catalog
    tab = pyiceberg_catalog.load_table(f"default.{table_name}")
    df = daft.read_table(f"{catalog_name}.default.{table_name}")
    df.collect()
    daft_pandas = df.to_pandas()
    iceberg_pandas = tab.scan().to_arrow().to_pandas()
    assert_df_equals(daft_pandas, iceberg_pandas, sort_key=[])


@pytest.mark.integration()
def test_daft_iceberg_table_renamed_filtered_collect_correct(local_iceberg_catalog):
    catalog_name, pyiceberg_catalog = local_iceberg_catalog
    tab = pyiceberg_catalog.load_table("default.test_table_rename")
    df = daft.read_table(f"{catalog_name}.default.test_table_rename")
    df = df.where(df["idx_renamed"] <= 1)
    daft_pandas = df.to_pandas()
    iceberg_pandas = tab.scan().to_arrow().to_pandas()
    iceberg_pandas = iceberg_pandas[iceberg_pandas["idx_renamed"] <= 1]
    assert_df_equals(daft_pandas, iceberg_pandas, sort_key=[])


@pytest.mark.integration()
def test_daft_iceberg_table_renamed_column_pushdown_collect_correct(local_iceberg_catalog):
    catalog_name, pyiceberg_catalog = local_iceberg_catalog
    tab = pyiceberg_catalog.load_table("default.test_table_rename")
    df = daft.read_table(f"{catalog_name}.default.test_table_rename")
    df = df.select("idx_renamed")
    daft_pandas = df.to_pandas()
    iceberg_pandas = tab.scan().to_arrow().to_pandas()
    iceberg_pandas = iceberg_pandas[["idx_renamed"]]
    assert_df_equals(daft_pandas, iceberg_pandas, sort_key=[])


@pytest.mark.integration()
def test_daft_iceberg_table_read_partition_column_identity(local_iceberg_catalog):
    catalog_name, pyiceberg_catalog = local_iceberg_catalog
    tab = pyiceberg_catalog.load_table("default.test_partitioned_by_identity")
    df = daft.read_table(f"{catalog_name}.default.test_partitioned_by_identity")
    df = df.select("ts", "number")
    daft_pandas = df.to_pandas()
    iceberg_pandas = tab.scan().to_arrow().to_pandas()
    iceberg_pandas = iceberg_pandas[["ts", "number"]]
    assert_df_equals(daft_pandas, iceberg_pandas, sort_key=[])


@pytest.mark.integration()
def test_daft_iceberg_table_read_partition_column_identity_filter(local_iceberg_catalog):
    catalog_name, pyiceberg_catalog = local_iceberg_catalog
    tab = pyiceberg_catalog.load_table("default.test_partitioned_by_identity")
    df = daft.read_table(f"{catalog_name}.default.test_partitioned_by_identity")
    df = df.where(df["number"] > 0)
    df = df.select("ts")
    daft_pandas = df.to_pandas()
    iceberg_pandas = tab.scan().to_arrow().to_pandas()
    iceberg_pandas = iceberg_pandas[iceberg_pandas["number"] > 0][["ts"]]
    assert_df_equals(daft_pandas, iceberg_pandas, sort_key=[])


@pytest.mark.skip(
    reason="Selecting just the identity-transformed partition key in an iceberg table is not yet supported. "
    "Issue: https://github.com/Eventual-Inc/Daft/issues/2129"
)
@pytest.mark.integration()
def test_daft_iceberg_table_read_partition_column_identity_filter_on_partkey(local_iceberg_catalog):
    catalog_name, pyiceberg_catalog = local_iceberg_catalog
    tab = pyiceberg_catalog.load_table("default.test_partitioned_by_identity")
    df = daft.read_table(f"{catalog_name}.default.test_partitioned_by_identity")
    df = df.select("ts")
    df = df.where(df["ts"] > datetime.date(2022, 3, 1))
    daft_pandas = df.to_pandas()
    iceberg_pandas = tab.scan().to_arrow().to_pandas()
    iceberg_pandas = iceberg_pandas[iceberg_pandas["ts"] > datetime.date(2022, 3, 1)][["ts"]]
    assert_df_equals(daft_pandas, iceberg_pandas, sort_key=[])


@pytest.mark.skip(
    reason="Selecting just the identity-transformed partition key in an iceberg table is not yet supported. "
    "Issue: https://github.com/Eventual-Inc/Daft/issues/2129"
)
@pytest.mark.integration()
def test_daft_iceberg_table_read_partition_column_identity_only(local_iceberg_catalog):
    catalog_name, pyiceberg_catalog = local_iceberg_catalog
    tab = pyiceberg_catalog.load_table("default.test_partitioned_by_identity")
    df = daft.read_table(f"{catalog_name}.default.test_partitioned_by_identity")
    df = df.select("ts")
    daft_pandas = df.to_pandas()
    iceberg_pandas = tab.scan().to_arrow().to_pandas()
    iceberg_pandas = iceberg_pandas[["ts"]]
    assert_df_equals(daft_pandas, iceberg_pandas, sort_key=[])


@pytest.mark.integration()
def test_daft_iceberg_table_read_partition_column_transformed(local_iceberg_catalog):
    catalog_name, pyiceberg_catalog = local_iceberg_catalog
    tab = pyiceberg_catalog.load_table("default.test_partitioned_by_bucket")
    df = daft.read_table(f"{catalog_name}.default.test_partitioned_by_bucket")
    df = df.select("number")
    daft_pandas = df.to_pandas()
    iceberg_pandas = tab.scan().to_arrow().to_pandas()
    iceberg_pandas = iceberg_pandas[["number"]]
    assert_df_equals(daft_pandas, iceberg_pandas, sort_key=[])


@pytest.mark.integration()
def test_daft_iceberg_table_read_table_snapshot(local_iceberg_catalog):
    _catalog_name, pyiceberg_catalog = local_iceberg_catalog
    tab = pyiceberg_catalog.load_table("default.test_snapshotting")

    snapshots = tab.history()
    assert len(snapshots) == 2

    for snapshot in snapshots:
        # TODO: Provide read_table API for reading iceberg with snapshot ID
        daft_pandas = daft.read_iceberg(tab, snapshot_id=snapshot.snapshot_id).to_pandas()
        iceberg_pandas = tab.scan(snapshot_id=snapshot.snapshot_id).to_pandas()
        assert_df_equals(daft_pandas, iceberg_pandas, sort_key=[])


@pytest.mark.integration()
@pytest.mark.parametrize("table_name", ["test_positional_mor_deletes", "test_positional_mor_double_deletes"])
def test_daft_iceberg_table_mor_limit_collect_correct(table_name, local_iceberg_catalog):
    catalog_name, pyiceberg_catalog = local_iceberg_catalog
    tab = pyiceberg_catalog.load_table(f"default.{table_name}")
    df = daft.read_table(f"{catalog_name}.default.{table_name}")
    df = df.limit(10)
    df.collect()
    daft_pandas = df.to_pandas()

    iceberg_arrow = tab.scan().to_arrow()
    iceberg_arrow = iceberg_arrow.slice(length=10)
    iceberg_pandas = iceberg_arrow.to_pandas()

    assert_df_equals(daft_pandas, iceberg_pandas, sort_key=[])


@pytest.mark.integration()
@pytest.mark.parametrize("table_name", ["test_positional_mor_deletes", "test_positional_mor_double_deletes"])
def test_daft_iceberg_table_mor_predicate_collect_correct(table_name, local_iceberg_catalog):
    catalog_name, pyiceberg_catalog = local_iceberg_catalog
    tab = pyiceberg_catalog.load_table(f"default.{table_name}")
    df = daft.read_table(f"{catalog_name}.default.{table_name}")
    df = df.where(df["number"] > 5)
    df.collect()
    daft_pandas = df.to_pandas()

    iceberg_arrow = tab.scan().to_arrow()
    iceberg_arrow = iceberg_arrow.filter(pc.field("number") > 5)
    iceberg_pandas = iceberg_arrow.to_pandas()

    assert_df_equals(daft_pandas, iceberg_pandas, sort_key=[])



from __future__ import annotations

import copy

import numpy as np
import pyarrow as pa
import pytest

from daft.datatype import DataType, get_super_ext_type
from daft.series import Series
from daft.utils import pyarrow_supports_fixed_shape_tensor
from tests.series import ARROW_FLOAT_TYPES, ARROW_INT_TYPES
from tests.utils import ANSI_ESCAPE

ARROW_VERSION = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric())
DaftExtension = get_super_ext_type()


@pytest.mark.parametrize("dtype", ARROW_INT_TYPES + ARROW_FLOAT_TYPES)
def test_tensor_roundtrip(dtype):
    np_dtype = dtype.to_pandas_dtype()
    data = [
        np.arange(8, dtype=np_dtype).reshape((2, 2, 2)),
        np.arange(8, 32, dtype=np_dtype).reshape((2, 2, 3, 2)),
        None,
    ]
    s = Series.from_pylist(data, pyobj="allow")

    daft_dtype = DataType.tensor(DataType.from_arrow_type(dtype))

    assert s.datatype() == daft_dtype

    # Test pylist roundtrip.
    back_dtype = DataType.python()
    back = s.cast(back_dtype)

    assert back.datatype() == back_dtype

    out = back.to_pylist()
    np.testing.assert_equal(out, data)

    # Test Arrow roundtrip.
    arrow_arr = s.to_arrow()

    assert isinstance(arrow_arr.type, DaftExtension)
    from_arrow = Series.from_arrow(arrow_arr)

    assert from_arrow.datatype() == s.datatype()
    np.testing.assert_equal(from_arrow.to_pylist(), s.to_pylist())

    s_copy = copy.deepcopy(s)
    assert s_copy.datatype() == s.datatype()
    np.testing.assert_equal(s_copy.to_pylist(), s.to_pylist())


@pytest.mark.parametrize("dtype", ARROW_INT_TYPES + ARROW_FLOAT_TYPES)
def test_fixed_shape_tensor_roundtrip(dtype):
    np_dtype = dtype.to_pandas_dtype()
    shape = (3, 2, 2)
    data = [
        np.arange(12, dtype=np_dtype).reshape(shape),
        np.arange(12, 24, dtype=np_dtype).reshape(shape),
        None,
    ]
    s = Series.from_pylist(data, pyobj="allow")

    target_dtype = DataType.tensor(DataType.from_arrow_type(dtype), shape)

    t = s.cast(target_dtype)

    assert t.datatype() == target_dtype

    # Test pylist roundtrip.
    back_dtype = DataType.python()
    back = t.cast(back_dtype)

    assert back.datatype() == back_dtype

    out = back.to_pylist()
    np.testing.assert_equal(out, data)

    # Test Arrow roundtrip.
    arrow_arr = t.to_arrow()

    if pyarrow_supports_fixed_shape_tensor():
        assert arrow_arr.type == pa.fixed_shape_tensor(dtype, shape)
    else:
        assert isinstance(arrow_arr.type, DaftExtension)
    from_arrow = Series.from_arrow(t.to_arrow())

    assert from_arrow.datatype() == t.datatype()
    np.testing.assert_equal(from_arrow.to_pylist(), t.to_pylist())

    if ARROW_VERSION >= (12, 0, 0):
        # Can't deepcopy pyarrow's fixed-shape tensor type.
        t_copy = t
    else:
        t_copy = copy.deepcopy(t)
    assert t_copy.datatype() == t.datatype()
    np.testing.assert_equal(t_copy.to_pylist(), t.to_pylist())


@pytest.mark.parametrize("dtype", ARROW_INT_TYPES + ARROW_FLOAT_TYPES)
@pytest.mark.parametrize("fixed_shape", [True, False])
def test_tensor_numpy_inference(dtype, fixed_shape):
    np_dtype = dtype.to_pandas_dtype()
    if fixed_shape:
        shape = (2, 2)
        arr = np.arange(np.prod(shape), dtype=np_dtype).reshape(shape)
        arrs = [arr, arr, None]
    else:
        shape1 = (2, 2)
        shape2 = (3, 3)
        arr1 = np.arange(np.prod(shape1), dtype=np_dtype).reshape(shape1)
        arr2 = np.arange(np.prod(shape1), np.prod(shape1) + np.prod(shape2), dtype=np_dtype).reshape(shape2)
        arrs = [arr1, arr2, None]
    s = Series.from_pylist(arrs, pyobj="allow")
    assert s.datatype() == DataType.tensor(DataType.from_arrow_type(dtype))
    out = s.to_pylist()
    np.testing.assert_equal(out, arrs)


def test_tensor_repr():
    arr = np.arange(np.prod((2, 2)), dtype=np.int64).reshape((2, 2))
    arrs = [arr, arr, None]
    s = Series.from_pylist(arrs, pyobj="allow")

    out_repr = ANSI_ESCAPE.sub("", repr(s))
    assert (
        out_repr.replace("\r", "")
        == """╭───────────────────────╮
│ list_series           │
│ ---                   │
│ Tensor(Int64)         │
╞═══════════════════════╡
│ <Tensor shape=(2, 2)> │
├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
│ <Tensor shape=(2, 2)> │
├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
│ None                  │
╰───────────────────────╯
"""
    )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from sortedcontainers import SortedDict

from daft.dependencies import pa

if TYPE_CHECKING:
    from daft.hudi.pyhudi.utils import FsFileMetadata


@dataclass(init=False)
class BaseFile:
    def __init__(self, fs_metadata: FsFileMetadata):
        self.metadata = fs_metadata
        file_name = fs_metadata.base_name
        self.file_name = file_name
        file_group_id, _, commit_time_ext = file_name.split("_")
        self.file_group_id = file_group_id
        self.commit_time = commit_time_ext.split(".")[0]

    @property
    def path(self) -> str:
        return self.metadata.path

    @property
    def size(self) -> int:
        return self.metadata.size

    @property
    def num_records(self) -> int:
        return self.metadata.num_records

    @property
    def schema(self) -> pa.Schema:
        return self.metadata.schema

    @property
    def colstats_schema(self) -> pa.Schema:
        return self.metadata.colstats_schema

    @property
    def min_values(self):
        return self.metadata.min_values

    @property
    def max_values(self):
        return self.metadata.max_values


@dataclass
class FileSlice:
    FILES_METADATA_SCHEMA = pa.schema(
        [
            ("path", pa.string()),
            ("size_bytes", pa.uint32()),
            ("num_records", pa.uint32()),
            ("partition_path", pa.string()),
        ]
    )

    file_group_id: str
    partition_path: str
    base_instant_time: str
    base_file: BaseFile

    @property
    def files_metadata(self):
        return self.base_file.path, self.base_file.size, self.base_file.num_records, self.partition_path

    @property
    def colstats_min_max(self):
        return self.base_file.min_values, self.base_file.max_values

    @property
    def colstats_schema(self):
        return self.base_file.colstats_schema


@dataclass
class FileGroup:
    file_group_id: str
    partition_path: str
    file_slices: SortedDict[str, FileSlice] = field(default_factory=SortedDict)

    def add_base_file(self, base_file: BaseFile):
        ct = base_file.commit_time
        if ct in self.file_slices:
            self.file_slices.get(ct).base_file = base_file
        else:
            self.file_slices[ct] = FileSlice(self.file_group_id, self.partition_path, ct, base_file)

    def get_latest_file_slice(self) -> FileSlice | None:
        if not self.file_slices:
            return None

        return self.file_slices.peekitem(-1)[1]



import itertools
from operator import add, mul, sub, truediv

import numpy as np
import pytest
from numpy.testing import assert_almost_equal

import daft
from daft import DataType

daft_signed_int_types = [
    DataType.int8(),
    DataType.int16(),
    DataType.int32(),
    DataType.int64(),
]

daft_numeric_types = daft_signed_int_types + [DataType.float32(), DataType.float64()]


@pytest.mark.parametrize(
    "op, ldtype, rdtype", itertools.product([add, mul, sub, truediv], daft_numeric_types, daft_numeric_types)
)
def test_math_tensors(op, ldtype, rdtype) -> None:
    np.random.seed(1)
    x = np.random.randint(0, 10, (12, 10, 1)).astype(ldtype.to_arrow_dtype().to_pandas_dtype())
    y = np.random.randint(0, 10, (12, 10, 1)).astype(rdtype.to_arrow_dtype().to_pandas_dtype())
    expected = op(x, y)

    df = daft.from_pydict({"x": x, "y": y})
    df = df.with_column("x", df["x"].cast(daft.DataType.tensor(ldtype, (10, 1))))
    df = df.with_column("y", df["y"].cast(daft.DataType.tensor(rdtype, (10, 1))))

    result = df.with_column("z", op(df["x"], df["y"])).collect()
    if op == truediv:
        assert result.schema()["z"].dtype == daft.DataType.tensor(daft.DataType.float64(), (10, 1))
    assert_almost_equal(result.to_pydict()["z"], expected)


@pytest.mark.parametrize(
    "op, ldtype, rdtype", itertools.product([add, mul, sub, truediv], daft_numeric_types, daft_numeric_types)
)
def test_math_tensors_with_literal(op, ldtype, rdtype) -> None:
    np.random.seed(1)
    x = np.random.randint(0, 10, (12, 10, 1)).astype(ldtype.to_arrow_dtype().to_pandas_dtype())
    y = np.random.randint(0, 10, (10, 1)).astype(rdtype.to_arrow_dtype().to_pandas_dtype())
    expected = op(x, y)

    df = daft.from_pydict({"x": x})
    df = df.with_column("x", df["x"].cast(daft.DataType.tensor(ldtype, (10, 1))))

    result = df.with_column("z", op(df["x"], daft.lit(y).cast(daft.DataType.tensor(rdtype, (10, 1))))).collect()
    if op == truediv:
        assert result.schema()["z"].dtype == daft.DataType.tensor(daft.DataType.float64(), (10, 1))

    assert_almost_equal(result.to_pydict()["z"], expected)


@pytest.mark.parametrize("op, ldtype", itertools.product([add, mul, sub, truediv], daft_numeric_types))
def test_math_tensors_with_null_literal(op, ldtype) -> None:
    np.random.seed(1)
    x = np.random.randint(0, 10, (12, 10, 1)).astype(ldtype.to_arrow_dtype().to_pandas_dtype())

    df = daft.from_pydict({"x": x})
    df = df.with_column("x", df["x"].cast(daft.DataType.tensor(ldtype, (10, 1))))

    result = df.with_column("z", op(df["x"], daft.lit(None).cast(daft.DataType.tensor(ldtype, (10, 1))))).collect()
    if op == truediv:
        assert result.schema()["z"].dtype == daft.DataType.tensor(daft.DataType.float64(), (10, 1))

    assert result.to_pydict()["z"] == [None] * 12
